@@ -22,7 +22,7 @@
  * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#ident "$Id: obj_hppa64.c 1.2 Wed, 28 Mar 2001 20:49:06 +1000 kaos $"
+#ident "$Id: obj_hppa64.c 1.3 Sun, 06 May 2001 15:09:08 +1000 kaos $"
 
 #include <string.h>
 #include <assert.h>
@@ -94,15 +94,15 @@ typedef struct _hppa64_file_t
  *     e8 20 d0 00     bve (r1)
  *     53 7b 00 30     ldd 18(dp),dp
  *
- * We need a different stub for millicode calls, which doesn't screw
- * dp:
+ * We need a different stub for millicode calls, which doesn't depend on
+ * or modify dp:
  *
- *     53 61 00 00     ldd 0(dp),r1
- *     			R_PARISC_LTOFF14R <.got entry offset from dp>
+ *     20 20 00 00     ldil 0,r1
+ *			R_PARISC_DIR21L <addr of kernels opd entry>
+ *     34 21 00 00     ldo 0(r1),r1
+ *			R_PARISC_DIR14R <addr of kernels opd entry>
  *     50 21 00 20     ldd 10(r1),r1
- *     e8 20 d0 00     bve (r1)
- *     08 00 02 40     nop
- *
+ *     e8 20 d0 02     bve,n (r1)
  */
 
 /* NOTE: to keep the code cleaner we make all stubs the same size.
@@ -120,10 +120,10 @@ unsigned char hppa64_stub_extern[] =
 
 unsigned char hppa64_stub_millicode[] =
 {
-       0x53, 0x61, 0x00, 0x00,
+       0x20, 0x20, 0x00, 0x00,
+       0x34, 0x21, 0x00, 0x00,
        0x50, 0x21, 0x00, 0x20,
-       0xe8, 0x20, 0xd0, 0x00,
-       0x08, 0x00, 0x02, 0x40,
+       0xe8, 0x20, 0xd0, 0x02,
 };
 
 /*======================================================================*/
@@ -575,15 +575,23 @@ arch_apply_relocation(struct obj_file *f,
 		     */
 		    unsigned char *stub;
 
-		    if (!strncmp(isym->root.name, "$$", 2))
+		    if (!strncmp(isym->root.name, "$$", 2)) {
 			stub = hppa64_stub_millicode;
-		    else
+			memcpy((Elf64_Addr *)(hfile->stub->contents + se->offset),
+					stub, SIZEOF_STUB);
+			v = (Elf64_Addr)isym->root.value;
+			ret = patch_21l(v, (Elf64_Word *)(hfile->stub->contents + se->offset));
+			if (ret == obj_reloc_ok)
+			    ret = patch_14r(v, (Elf64_Word *)(hfile->stub->contents + se->offset + 4));
+		    }
+		    else {
 			stub = hppa64_stub_extern;
+			memcpy((Elf64_Addr *)(hfile->stub->contents + se->offset),
+					stub, SIZEOF_STUB);
+			v = (Elf64_Addr)(hfile->got->header.sh_addr + ge->offset) - gp;
+			ret = patch_14r2(v, (Elf64_Word *)(hfile->stub->contents + se->offset));
+		    }
 		    se->reloc_done = TRUE;
-		    memcpy((Elf64_Addr *)(hfile->stub->contents + se->offset),
-	    			stub, SIZEOF_STUB);
-		    v = (Elf64_Addr)(hfile->got->header.sh_addr + ge->offset) - gp;
-		    ret = patch_14r2(v, (Elf64_Word *)(hfile->stub->contents + se->offset));
 		}
 		v = hfile->stub->header.sh_addr + se->offset;
 	    }
